﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Text;
using System.Globalization;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 달력 컨트롤
    /// </summary>
    public partial class CalendarControl : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 컨트롤 마진
        /// </summary>
        private const int CONTROL_MARGIN = 3;

        /// <summary>
        /// 헤더 높이
        /// </summary>
        private const int HEADER_HEIGHT = 25;

        /// <summary>
        /// 달력 정보
        /// </summary>
        private CalendarInformation calendarInformation = null;

        /// <summary>
        /// 요일 딕셔너리
        /// </summary>
        private Dictionary<int, WeekdayItem> weekdayDictionary = new Dictionary<int, WeekdayItem>();

        /// <summary>
        /// 일 딕셔너리
        /// </summary>
        private Dictionary<CalendarItem, DayItem> dayDictionary = new Dictionary<CalendarItem, DayItem>();

        /// <summary>
        /// 요일 제목 폰트
        /// </summary>
        private Font weekdayTitleFont;
        
        /// <summary>
        /// 일 제목 당월 폰트
        /// </summary>
        private Font dayTitleCurrentMonthFont;

        /// <summary>
        /// 일 제목 비당월 폰트
        /// </summary>
        private Font dayTitleNonCurrentMonthFont;

        /// <summary>
        /// 일 제목 당일 폰트
        /// </summary>
        private Font dayTitleTodayFont;

        /// <summary>
        /// 내용 폰트
        /// </summary>
        private Font contentFont;

        /// <summary>
        /// 요일 제목 문자열 포맷
        /// </summary>
        private StringFormat weekdayTitleStringFormat;

        /// <summary>
        /// 내용 문자열 포맷
        /// </summary>
        private StringFormat contentStringFormat;

        /// <summary>
        /// 요일명 배열
        /// </summary>
        private string[] dayNameArray;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 달력 정보 - CalendarInformation

        /// <summary>
        /// 달력 정보
        /// </summary>
        public CalendarInformation CalendarInformation
        {
            get
            {
                return this.calendarInformation;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - CalendarControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public CalendarControl()
        {
            SetDoubleBuffered(this, true);

            InitializeComponent();

            this.weekdayTitleFont            = new Font("나눔고딕코딩", 12f);
            this.dayTitleCurrentMonthFont    = new Font("나눔고딕코딩", 12f, FontStyle.Bold);
            this.dayTitleNonCurrentMonthFont = new Font("나눔고딕코딩", 12f);
            this.dayTitleTodayFont           = new Font("나눔고딕코딩", 16f, FontStyle.Bold);
            this.contentFont                 = new Font("나눔고딕코딩", 12f);

            this.weekdayTitleStringFormat = new StringFormat();

            this.weekdayTitleStringFormat.Alignment     = StringAlignment.Center;
            this.weekdayTitleStringFormat.LineAlignment = StringAlignment.Center;

            this.contentStringFormat = new StringFormat();

            this.contentStringFormat.Alignment     = StringAlignment.Near;
            this.contentStringFormat.LineAlignment = StringAlignment.Center;
            
            this.dayNameArray = DateTimeFormatInfo.CurrentInfo.DayNames;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 달력 정보 설정하기 - SetCalendarInformation(calendarInformation)

        /// <summary>
        /// 달력 정보 설정하기
        /// </summary>
        /// <param name="calendarInformation">연도</param>
        public void SetCalendarInformation(CalendarInformation calendarInformation)
        {
            this.calendarInformation = calendarInformation;
            
            Calculate();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 크기 변경시 처리하기 - OnSizeChanged(e)

        /// <summary>
        /// 크기 변경시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            Calculate();
        }

        #endregion
        #region 페인트시 처리하기 - OnPaint(e)

        /// <summary>
        /// 페인트시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            Graphics graphics = e.Graphics;

            #region 그래픽스 객체 속성을 설정한다.

            graphics.InterpolationMode  = InterpolationMode.HighQualityBicubic;
            graphics.CompositingQuality = CompositingQuality.HighQuality;
            graphics.SmoothingMode      = SmoothingMode.HighQuality;
            graphics.TextRenderingHint  = TextRenderingHint.ClearTypeGridFit;
            graphics.PixelOffsetMode    = PixelOffsetMode.HighQuality;

            #endregion
            #region 컨트롤 배경을 칠한다.

            graphics.Clear(Color.White);

            #endregion
            #region 컨트롤 테두리 선을 그린다.

            graphics.DrawRectangle(Pens.Black, 0, 0, Width, Height);

            #endregion
            #region 요일 제목을 그린다.

            foreach(KeyValuePair<int, WeekdayItem> keyValuePair in this.weekdayDictionary)
            {
                int         weekDay     = keyValuePair.Key;
                WeekdayItem weekDayItem = keyValuePair.Value;

                graphics.DrawString(weekDayItem.WeekdayTitle, this.weekdayTitleFont, Brushes.Black, weekDayItem.Rectangle, this.weekdayTitleStringFormat); 
            }

            #endregion

            foreach(KeyValuePair<CalendarItem, DayItem> keyValuePair in this.dayDictionary)
            {
                CalendarItem calendarItem = keyValuePair.Key;
                DayItem      dayItem      = keyValuePair.Value;

                graphics.SetClip(dayItem.Rectangle);

                #region 셀 배경을 칠한다.

                graphics.FillRectangle(dayItem.BackgroundBrush, dayItem.Rectangle);

                #endregion
                #region 셀 테두리 선을 그린다.

                graphics.DrawRectangle(dayItem.BorderPen , dayItem.Rectangle);

                #endregion
                #region 일 제목을 그린다.

                graphics.DrawString(dayItem.DayTitle, dayItem.DayTitleFont, Brushes.Black, dayItem.X + CONTROL_MARGIN, dayItem.Y + CONTROL_MARGIN);

                #endregion

                for(int i = 0; i < calendarItem.ContentList.Count; i++)
                {
                    string content = calendarItem.ContentList[i];

                    graphics.DrawString(content, this.contentFont, Brushes.Black, dayItem.X + CONTROL_MARGIN, dayItem.Y + CONTROL_MARGIN + HEADER_HEIGHT + i * 20);
                }
                

                graphics.ResetClip();
            }
        }

        #endregion
        #region 마우스 더블 클릭시 처리하기 - OnMouseDoubleClick(e)

        /// <summary>
        /// 마우스 더블 클릭시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            base.OnMouseDoubleClick(e);

            foreach(KeyValuePair<CalendarItem, DayItem> keyValuePair in this.dayDictionary)
            {
                CalendarItem calendarItem = keyValuePair.Key;
                DayItem      dayItem      = keyValuePair.Value;

                if(dayItem.Rectangle.Contains(e.Location))
                {
                    StringBuilder stringBuilder = new StringBuilder();

                    for(int i = 0; i < calendarItem.ContentList.Count; i++)
                    {
                        stringBuilder.AppendLine(calendarItem.ContentList[i]);
                    }

                    InputContentPopupForm popupForm = new InputContentPopupForm();

                    popupForm.DateLabel = calendarItem.Date.ToString("yyyy-MM-dd");
                    popupForm.Content   = stringBuilder.ToString().Trim();

                    if(popupForm.ShowDialog(this) == DialogResult.OK)
                    {
                        string[] contentArray = popupForm.Content.Split('\n');

                        calendarItem.ContentList.Clear();

                        foreach(string content in contentArray)
                        {
                            if(!string.IsNullOrEmpty(content))
                            {
                                calendarItem.ContentList.Add(content);
                            }
                        }

                        Invalidate();
                    }
                }
            }                        
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 더블 버퍼링 설정하기 - SetDoubleBuffered(control, doubleBuffered)

        /// <summary>
        /// 더블 버퍼링 설정하기
        /// </summary>
        /// <param name="control">컨트롤</param>
        /// <param name="doubleBuffered">더블 버퍼링 여부</param>
        private void SetDoubleBuffered(Control control, bool doubleBuffered = true)
        {
            PropertyInfo propertyInfo = typeof(Control).GetProperty("DoubleBuffered", BindingFlags.Instance | BindingFlags.NonPublic);
    
            propertyInfo.SetValue(control, doubleBuffered, null);
        }

        #endregion
        #region 계산하기 - Calculate()

        /// <summary>
        /// 계산하기
        /// </summary>
        private void Calculate()
        {
            this.weekdayDictionary.Clear();
            this.dayDictionary.Clear();

            if(Width == 0 || Height == 0)
            {
                return;
            }

            if(this.calendarInformation == null)
            {
                return;
            }

            int rowCount    = this.calendarInformation.RowCount;
            int columnCount = this.calendarInformation.ColumnCount;

            int rowHeight   = (int)((double)(Height - CONTROL_MARGIN - HEADER_HEIGHT) / rowCount   );
            int columnWidth = (int)((double)(Width  - CONTROL_MARGIN                ) / columnCount);

            for(int i = 0; i < 7; i++)
            {
                WeekdayItem item = new WeekdayItem();

                item.X         = i * columnWidth + CONTROL_MARGIN;
                item.Y         = CONTROL_MARGIN;
                item.Width     = columnWidth - CONTROL_MARGIN;
                item.Height    = HEADER_HEIGHT;
                item.Rectangle = new Rectangle(item.X, item.Y, item.Width, item.Height);

                if(columnWidth > 60)
                {
                    item.WeekdayTitle = this.dayNameArray[i];
                }
                else
                {
                    item.WeekdayTitle = this.dayNameArray[i].Replace("요일", string.Empty);
                }

                this.weekdayDictionary.Add(i, item);
            }

            foreach(CalendarItem calendarItem in this.calendarInformation.ItemList)
            {
                DayItem cellItem = new DayItem();

                cellItem.X         = calendarItem.ColumnIndex * columnWidth + CONTROL_MARGIN;
                cellItem.Y         = calendarItem.RowIndex    * rowHeight   + CONTROL_MARGIN + HEADER_HEIGHT;
                cellItem.Width     = columnWidth - CONTROL_MARGIN;
                cellItem.Height    = rowHeight   - CONTROL_MARGIN;
                cellItem.Rectangle = new Rectangle(cellItem.X, cellItem.Y, cellItem.Width, cellItem.Height);

                if(calendarItem.IsPreviousMonth || calendarItem.IsNexMonth)
                {
                    cellItem.BackgroundBrush = Brushes.WhiteSmoke;
                    cellItem.BorderPen       = Pens.DarkGray;

                    cellItem.DayTitleFont = this.dayTitleNonCurrentMonthFont;

                    if(columnWidth > 60)
                    {
                        cellItem.DayTitle = string.Format("{0}/{1}", calendarItem.Date.Month, calendarItem.Date.Day);
                    }
                    else
                    {
                        cellItem.DayTitle = calendarItem.Date.Day.ToString();
                    }
                }
                else
                {
                    if(calendarItem.IsToday)
                    {
                        if(calendarItem.ColumnIndex == 0 || calendarItem.ColumnIndex == 6)
                        {
                            cellItem.BackgroundBrush = Brushes.LavenderBlush;
                            cellItem.BorderPen       = new Pen(Brushes.DarkGray, 4f);
                        }
                        else
                        {
                            cellItem.BackgroundBrush = Brushes.White;
                            cellItem.BorderPen       = new Pen(Brushes.DarkGray, 4f);
                        }
                    }
                    else
                    {
                        if(calendarItem.ColumnIndex == 0 || calendarItem.ColumnIndex == 6)
                        {
                            cellItem.BackgroundBrush = Brushes.LavenderBlush;
                            cellItem.BorderPen       = Pens.DarkGray;
                        }
                        else
                        {
                            cellItem.BackgroundBrush = Brushes.White;
                            cellItem.BorderPen       = Pens.DarkGray;
                        }
                    }

                    if(calendarItem.IsToday)
                    {
                        cellItem.DayTitleFont = this.dayTitleTodayFont;
                        cellItem.DayTitle     = calendarItem.Date.Day.ToString();
                    }
                    else
                    {
                        cellItem.DayTitleFont = this.dayTitleCurrentMonthFont;
                        cellItem.DayTitle     = calendarItem.Date.Day.ToString();
                    }
                }

                this.dayDictionary.Add(calendarItem, cellItem);
            }

            Invalidate();
        }

        #endregion
    }
}